<?php

namespace SuperRhino\Utils\Redis;

/**
 * redis令牌桶
 */
class Permits {

    /**
     * 令牌桶
     * @param string $key
     * @param int $permits 获取令牌数量
     * @param int $maxPermits 令牌桶容量
     * @param float $rate 令牌生成qps
     * @return bool
     */
    public static function permits(string $key, int $permits, int $maxPermits, float $rate) {
        if (empty($key)) {
            return false;
        }

        $redis = Redis::getRedis();
        $curr_mill_second = intval(microtime(true) * 1000);

        /*
         * 令牌桶刚刚创建，上一次获取令牌的毫秒数为空
         * 根据和上一次向桶里添加令牌的时间和当前时间差，触发式往桶里添加令牌, 并且更新上一次向桶里添加令牌的时间
         * 如果向桶里添加的令牌数不足一个，则不更新上一次向桶里添加令牌的时间
         */
        $luaScript = <<<LUA
local rate_limit_info = redis.pcall("HMGET", KEYS[1], "last_mill_second", "curr_permits")
local last_mill_second = rate_limit_info[1]
local curr_permits = tonumber(rate_limit_info[2])

local local_curr_permits = ARGV[3];

if (type(last_mill_second) ~= 'boolean' and last_mill_second ~= false and last_mill_second ~= nil) then
    --- 添加令牌数量
    local reverse_permits = math.floor(((ARGV[2] - last_mill_second) / 1000) * ARGV[4])
    --- 总令牌
    local expect_curr_permits = reverse_permits + curr_permits;
    --- 限制总令牌数
    local_curr_permits = math.min(expect_curr_permits, ARGV[3]);

    --- 大于0表示不是第一次获取令牌，也没有向桶里添加令牌
    if (reverse_permits > 0) then
        redis.pcall("HSET", KEYS[1], "last_mill_second", ARGV[2])
    end
else
    redis.pcall("HSET", KEYS[1], "last_mill_second", ARGV[2])
end


local result = -1
if (local_curr_permits - ARGV[1] >= 0) then
    result = 1
    redis.pcall("HSET", KEYS[1], "curr_permits", local_curr_permits - ARGV[1])
else
    redis.pcall("HSET", KEYS[1], "curr_permits", local_curr_permits)
end
redis.call('EXPIRE', KEYS[1], math.ceil(ARGV[3]/ARGV[4]))
return result
LUA;
        $result = $redis->eval($luaScript, ['permits_' . $key, $permits, $curr_mill_second, $maxPermits, $rate], 1);
        return $result > 0;
    }
}